namespace nkmidi
{
	using System;
	using System.Collections;
	using System.Runtime.InteropServices;
	using System.Threading;

	using nkmidi.Meta;

	/// <summary>
	/// Player: MIDIV[PXĐvC[NX
	/// </summary>
	public class Player
	{
		// import from winmm.dll and mmsystem.h
		[DllImport("winmm.dll")]
		static extern int midiOutClose(int hMidiOut);
		[DllImport("winmm.dll")]
		static extern int midiOutOpen(ref int lphMidiOut, int uDeviceID, int dwCallback, int dwInstance, int dwFlags);
		[DllImport("winmm.dll")]
		static extern int midiOutReset(int hMidiOut);
		[DllImport("winmm.dll")]
		static extern int midiOutShortMsg(int hMidiOut, int dwMsg);

		const int MMSYSERR_NOERROR = 0;
		const int MIDI_MAPPER = -1;



		// member variables

		private string file;
		private Sequence seq;
		private Thread thPlay;
		private int tempo;
		private int hMidiOut;
		private int deviceID;
		private int sleepTimeUnit; // miliseconds
		private bool flagPlaying;
		private ArrayList handlers;
		private NotifyCallback endOfSequenceCallback;



		// methods

		public Player()
		{
			file = "";
			seq = null;
			thPlay = null;
			this.Tempo = 0;
			hMidiOut = 0;
			deviceID = MIDI_MAPPER;
			sleepTimeUnit = 20;
			flagPlaying = false;
			handlers = new ArrayList();
			endOfSequenceCallback = null;
		}

		public void load(string filename) 
		{
			if (flagPlaying)
			{
				throw( new Exception("load(): now playing") ); // TODO
			}

			file = filename;

			SMF smf = new SMF(file);
			seq = new Sequence(smf);
			this.Tempo = seq.StartingTempo;
		}

		public void play()
		{
			this.stop();

			int error_code = midiOutOpen(ref hMidiOut, deviceID, 0, 0, 0);

			if (error_code != MMSYSERR_NOERROR) 
			{
				throw( new Exception("midiOutOpen(): an error occured") );
			}

			thPlay = new Thread( new ThreadStart(this.startPlay) );
			thPlay.Start();
		}

		public void play(string filename)
		{
			this.stop();

			this.load(filename);
			this.play();
		}

		private void startPlay()
		{
			try
			{
				flagPlaying = true;

				MIDITrack[] tracks = seq.Tracks;
				int num_tracks = tracks.Length;
				int[] remain = new int[num_tracks]; // remain time to the waiting event of each track
				int[] waiting = new int[num_tracks];   // index of the waiting (next) event of each track
				int time_base = seq.TimeBase;

				// initialization for remain[] and waiting[]
				int current_tempo = this.Tempo;
				int tempo_msec = (int)( (double)60 * 1000 / current_tempo ); // miliseconds
				for (int i=0; i<num_tracks; i++) 
				{
					waiting[i] = 0;

					MIDIEvent ev = ( tracks[i].getEvents() )[ waiting[i] ];
					remain[i] = (int)( ((double)ev.DeltaTime / time_base) * (double)tempo_msec );
				}

				while (true)
				{
					// pick up the event which has lowest remain time
					int picked = 0; // track number which is picked up
					bool flagNoEvents = true; // check if there's an event to be processed
					for (int i=0; i<num_tracks; i++)
					{
						if (waiting[i] >= 0)
						{
							if ( waiting[picked] < 0 || remain[i] < remain[picked] )
							{
								picked = i;
							}
							flagNoEvents = false;
						}
					}
					if (flagNoEvents)
					{
						// break from while loop if there's no event
						break;
					}

					// wait until the next event
					while ( remain[picked] > 0 ) 
					{
						int sleepTime = (sleepTimeUnit < remain[picked] ? sleepTimeUnit : remain[picked]);
						Thread.Sleep(sleepTime);

						// check if there's a tempo change
						double ratio = 1.0;
						if ( current_tempo != this.Tempo )
						{
							ratio = (double)this.Tempo / current_tempo;
							current_tempo = this.Tempo;
							tempo_msec = (int)( (double)60 * 1000 / current_tempo ); // miliseconds
						}

						// subtract slept time from remain time
						for (int i=0; i<num_tracks; i++)
						{
							if (waiting[i] < 0)
							{
								continue;
							}

							remain[i] -= sleepTime;
							remain[i] = (int)(remain[i] / ratio);
						}
					}

					MIDIEvent[] events = tracks[picked].getEvents();
					MIDIEvent ev = events[ waiting[picked] ];

					// process the event
					this.runEvent(ev);

					// set to next waiting event
					waiting[picked]++;
					if (waiting[picked] >= events.Length)
					{
						waiting[picked] = -1;
					}
					else
					{
						MIDIEvent next_event = events[ waiting[picked] ];
						remain[picked] = (int)( ((double)next_event.DeltaTime / time_base) * (double)tempo_msec );
					}
				}
			}
			catch (ThreadAbortException) 
			{
			}
			finally
			{
				this.stopPlay();
				if ( endOfSequenceCallback != null )
				{
					endOfSequenceCallback();
				}
			}
		}

		private void runEvent(MIDIEvent ev)
		{
			if ( this.invokeEventHandler(ev, EventHandler.ANY_EVENT) )
			{
				return;
			}

			if ( ev.IsMetaEvent ) 
			{
				// Meta Event
				if ( this.invokeEventHandler(ev, EventHandler.META_EVENT) )
				{
					return;
				}

				MetaEvent meta_event = ev.MetaEvent;
				if (meta_event.EventType == MetaEvent.EVENTTYPE_SET_TEMPO)
				{
					// Set tempo
					MetaSetTempo mev = (MetaSetTempo)meta_event;
					int tempo_usec = mev.Tempo;
					this.Tempo = (int)(60 * 1000000 / (double)tempo_usec);
				}

				this.invokeMetaEventHandler(meta_event, meta_event.EventType);
			}
			else if ( ev.IsSysexEvent )
			{
				// SysEx Event
				if ( this.invokeEventHandler(ev, EventHandler.SYSEX_EVENT) )
				{
					return;
				}

				// TODO
			}
			else
			{
				// MIDI Channel Event
				if ( this.invokeEventHandler(ev, EventHandler.MIDI_CHANNEL_EVENT) )
				{
					return;
				}

				byte[] bytes = ev.RawData;
				int msg = 0;

				for (int i=bytes.Length-1; i>=0; i--)
				{
					msg = (msg << 8);
					msg += bytes[i];
				}
				midiOutShortMsg(hMidiOut, msg);
			}
		}

		private bool invokeEventHandler(MIDIEvent ev, int event_type)
		{
			bool interception = false;
			for (int i=0; i<handlers.Count; i++)
			{
				if ( handlers[i].GetType() == Type.GetType("nkmidi.EventHandler") )
				{
					// the object is an EventHandler
					EventHandler handler = (EventHandler)handlers[i];
					if ( handler.EventType == event_type && handler.Callback != null )
					{
						interception = handler.Callback(ev);
					}
				}
			}

			// return the last value of the callbacks
			return interception;
		}

		private void invokeMetaEventHandler(MetaEvent ev, int event_type)
		{
			for (int i=0; i<handlers.Count; i++)
			{
				if ( handlers[i].GetType() == Type.GetType("nkmidi.MetaEventHandler") )
				{
					// the object is a MetaEventHandler
					MetaEventHandler handler = (MetaEventHandler)handlers[i];
					if ( handler.MetaEventType == event_type && handler.Callback != null )
					{
						handler.Callback(ev);
					}
				}
			}
		}

		public void stop()
		{
			if ( !flagPlaying )
			{
				return;
			}

			thPlay.Abort();
		}

		private void stopPlay()
		{
			if ( !flagPlaying )
			{
				return;
			}

			midiOutReset(hMidiOut);
			midiOutClose(hMidiOut);

			flagPlaying = false;
		}



		// handlers

		public void addHandler(EventHandler handler)
		{
			handlers.Add(handler);
		}

		public void clearHandlers()
		{
			handlers.Clear();
		}

		public void removeHandler(EventHandler handler)
		{
			handlers.Remove(handler);
		}


		// accessors and properties

		public int Tempo
		{
			get { return tempo; }
			set { tempo = value; }
		}

		public int DeviceID
		{
			get { return deviceID; }
			set { deviceID = value; }
		}

		public bool IsPlaying
		{
			get { return flagPlaying; }
		}

		public Sequence Sequence
		{
			get { return seq; }
		}

		public NotifyCallback EndOfSequenceCallback
		{
			set { endOfSequenceCallback = value; }
		}
	}

	public delegate void NotifyCallback();
	public delegate bool EventCallback(MIDIEvent ev);
	public delegate void MetaEventCallback(MetaEvent mev);

	public class EventHandler
	{
		private int eventType;
		private EventCallback callback;

		public const int ANY_EVENT = 0;
		public const int MIDI_CHANNEL_EVENT = 1;
		public const int SYSEX_EVENT = 2;
		public const int META_EVENT = 3;

		public EventHandler()
		{
			this.EventType = ANY_EVENT;
			this.Callback = null;  
		}

		public int EventType
		{
			get { return eventType; }
			set { eventType = value; }
		}

		public EventCallback Callback
		{
			get { return callback; }
			set { callback = value; }
		}
	}

	public class MetaEventHandler : EventHandler
	{
		private int metaEventType;
		private MetaEventCallback metaCallback;

		public MetaEventHandler(int meta_event_type)
		{
			this.EventType = EventHandler.META_EVENT;
			this.MetaEventType = meta_event_type;
			this.Callback = null;
		}

		public int MetaEventType
		{
			get { return metaEventType; }
			set { metaEventType = value; }
		}

		new public MetaEventCallback Callback
		{
			get { return metaCallback; }
			set { metaCallback = value; }
		}
	}
}
