namespace nkmidi
{
	using System;
	using System.IO;
	using System.Text;

	using nkmidi.Meta;

	/// <summary>
	/// SMF: SMF`MIDIf[^
	/// </summary>
	public class SMF
	{
		private string file;
		private byte[] bytes;
		private MIDITrack[] tracks;

		private string headerChunkType;
		private uint headerDataLength;
		private uint formatNumber;
		private uint numberOfTracks;
		private uint timeBase;

		Encoding asciiEncoding;
		Encoding localEncoding;

		public SMF() 
		{
			file = "";
			bytes = null;
			tracks = null;

			this.HeaderChunkType = "MThd";
			this.HeaderDataLength = 0;
			formatNumber = 0;
			numberOfTracks = 0;
			timeBase = 0;

			asciiEncoding = Encoding.ASCII;
			localEncoding = Encoding.GetEncoding("Shift_JIS"); // TODO: Generalization
		}

		public SMF(string filename) : this()
		{
			file = filename;
		}

		public void read() 
		{
			FileStream fs = new FileStream(file, FileMode.Open, FileAccess.Read);
			bytes = new byte[fs.Length];
			fs.Read(bytes, 0, bytes.Length);
			fs.Close();

			int p = 0; // index (pointer) of the data

			//
			// get header chunk type
			//
			// Chunk type is 4-byte text and must be "MThd"
			//

			// TODO: throw exception
			string header_chunk_type = asciiEncoding.GetString(bytes, p, 4);
			this.HeaderChunkType = header_chunk_type;
			p += 4;


			//
			// get header data length
			//
			// Data length is 4-byte uint and will be 6

			// TODO: throw exception
			uint header_data_length = this.convertToUIntFromBytes(bytes, p, 4);
			this.HeaderDataLength = header_data_length;
			p += 4;

		
			//
			// get format number
			//
			// SMF format number is 2-byte uint and will be 0 or 1
			// Format 0: single track
			// Format 1: multi tracks
			// Format 2: not supported in this library
			formatNumber = this.convertToUIntFromBytes(bytes, p, 2);
			p += 2;


			//
			// get number of tracks
			//
			// Number of tracks is 2-byte uint

			// TODO: throw if not greater than or equal to 1
			numberOfTracks = this.convertToUIntFromBytes(bytes, p, 2);
			p += 2;

			//
			// get time base
			//
			// Time base is 2-byte uint
			timeBase = this.convertToUIntFromBytes(bytes, p, 2);
			p += 2;

			tracks = new MIDITrack[numberOfTracks];
			for (int i=0; i<numberOfTracks; i++) 
			{
				tracks[i] = new MIDITrack();

				//
				// get track chunk type
				//
				// Chunk type is 4-byte text and must be "MTrk"
				string track_chunk_type = asciiEncoding.GetString(bytes, p, 4);
				p += 4;

				//
				// get track data length
				//
				uint track_data_length = this.convertToUIntFromBytes(bytes, p, 4);
				p += 4;

				// process the track data
				uint position_of_next_track = (uint)(p + track_data_length); // position of next track begins
				byte previous_status_byte = 0;
				while (p < position_of_next_track) 
				{
					MIDIEvent ev = new MIDIEvent();

					//
					// get the delta time
					//
					// delta time is recorded as variable-length value
					int processed;
					uint delta_time = this.convertVariableLengthValue(bytes, p, out processed);
					p += processed;

					ev.DeltaTime = delta_time;

					// 
					// check the running status
					//					
					// status byte must be greater than or equal to 0x80 (the first bit is 1)
					byte status_byte;
					if ( (previous_status_byte > 0) && ((bytes[p] & 0x80) == 0) )
					{
						// running status: status byte is omitted - use previous status byte
						status_byte = previous_status_byte;
					}
					else
					{
						status_byte = bytes[p];
						p++;
					}
					previous_status_byte = status_byte;

					if (0x80 <= status_byte && status_byte < 0xF0) 
					{
						// MIDI Channel Events

						// data length for each message types
						int[] bytes_of_messages = new int[] {
																 3, // 0x8n: note off
																 3, // 0x9n: note on
																 3, // 0xAn: polyphonic key pressure (note aftertouch)
																 3, // 0xBn: control change
																 2, // 0xCn: program change
																 2, // 0xDn: channel pressure (aftertouch)
																 3  // 0xEn: pitch bend (pitch wheel) change
															 };

						int msglen = bytes_of_messages[(status_byte >> 4) - 8];
						ev.RawData = this.copyBytesWithLeadingByte(bytes, p, msglen - 1, status_byte);
						p += (msglen - 1);
					}
					else if (status_byte == 0xF0 || status_byte == 0xF7)
					{
						// SysEx Events
						//
						// 0xF0: used for system exclusive message
						// 0xF7: used for system exclusive message,
						//       common message(0xF1, 0xF2, 0xF3, 0xF6, 0xF7),
						//       realtime message(0xF8, 0xFA, 0xFB, 0xFC, 0xFE, 0xFF)

						// data length is recorded as variable-length value
						int ev_processed;
						int ev_data_length = (int)this.convertVariableLengthValue(bytes, p, out ev_processed);
						p += ev_processed;

						if (status_byte == 0xF0) 
						{
							// include the status byte (0xF0) in the beginning
							ev.RawData = this.copyBytesWithLeadingByte(bytes, p, ev_data_length, status_byte);
						}
						else 
						{
							ev.RawData = this.copyBytes(bytes, p, ev_data_length);
						}
						p += ev_data_length;
					}
					else if (status_byte == 0xFF) 
					{
						// Meta Events
						byte event_type = bytes[p];
						p++;

						// data length is recorded as variable-length value
						int ev_processed;
						int ev_data_length = (int)this.convertVariableLengthValue(bytes, p, out ev_processed);
						p += ev_processed;

						byte[] meta_data = this.copyBytes(bytes, p, ev_data_length);
						ev.RawData = meta_data;
						p += ev_data_length;

						MetaEvent meta_event = null;
						int mp = 0; // pointer for meta data
						if (event_type == 0x00)
						{
							// Sequence Number
							MetaSequenceNumber mev = new MetaSequenceNumber();
							uint seq_number = this.convertToUIntFromBytes(meta_data, mp, 2);
							mev.SequenceNumber = (int)seq_number;
							meta_event = mev;
						}
						else if (event_type == 0x01)
						{
							// Text Event
							MetaText mev = new MetaText();
							string str = localEncoding.GetString(meta_data);
							mev.Text = str;
							meta_event = mev;
						}
						else if (event_type == 0x02)
						{
							// Copyright Notice
							MetaCopyright mev = new MetaCopyright();
							string str = localEncoding.GetString(meta_data);
							mev.Copyright = str;
							meta_event = mev;
						}
						else if (event_type == 0x03)
						{
							// Sequence Name or Track Name
							MetaSequenceOrTrackName mev = new MetaSequenceOrTrackName();
							string str = localEncoding.GetString(meta_data);
							mev.Name = str;
							meta_event = mev;
						}
						else if (event_type == 0x04)
						{
							// Instrument Name
							MetaInstrumentName mev = new MetaInstrumentName();
							string str = localEncoding.GetString(meta_data);
							mev.InstrumentName = str;
							meta_event = mev;
						}
						else if (event_type == 0x05)
						{
							// Lyrics
							MetaLyrics mev = new MetaLyrics();
							string str = localEncoding.GetString(meta_data);
							mev.Lyrics = str;
							meta_event = mev;
						}
						else if (event_type == 0x06)
						{
							// Marker
							MetaMarker mev = new MetaMarker();
							string str = localEncoding.GetString(meta_data);
							mev.Marker = str;
							meta_event = mev;
						}
						else if (event_type == 0x07)
						{
							// Cue Point
							MetaCuePoint mev = new MetaCuePoint();
							string str = localEncoding.GetString(meta_data);
							mev.CuePoint = str;
							meta_event = mev;
						}
						else if (event_type == 0x20)
						{
							// MIDI Channel Prefix
							MetaMIDIChannelPrefix mev = new MetaMIDIChannelPrefix();
							int channel = meta_data[mp];
							mev.ChannelPrefix = channel;
							meta_event = mev;
						}
						else if (event_type == 0x21)
						{
							// MIDI Port
							MetaMIDIPort mev = new MetaMIDIPort();
							int port = meta_data[mp];
							mev.Port = port;
							meta_event = mev;
						}
						else if (event_type == 0x2F)
						{
							// End of Track
							MetaEndOfTrack mev = new MetaEndOfTrack();
							meta_event = mev;
						}
						else if (event_type == 0x51)
						{
							// Set Tempo
							MetaSetTempo mev = new MetaSetTempo();
							uint tempo_usec = this.convertToUIntFromBytes(meta_data, mp, 3);
							mev.Tempo = (int)tempo_usec;
							meta_event = mev;
						}
						else if (event_type == 0x54)
						{
							// SMPTE Offset
							MetaSMPTEOffset mev = new MetaSMPTEOffset();

							int hr_byte = meta_data[mp];
							int frame_type = ( (hr_byte & 0x60) >> 5 ); // 0rr00000
							mev.FrameType = frame_type;
							int hours = ( hr_byte & 0x1F ); // 000hhhhh
							mev.Hours = hours;
							mp++;
							int min = meta_data[mp];
							mev.Minutes = min;
							mp++;
							int sec = meta_data[mp];
							mev.Seconds = sec;
							mp++;
							int frames = meta_data[mp];
							mev.Frames = frames;
							mp++;
							int frac_frames = meta_data[mp];
							mev.FractionalFrames = frac_frames;
							mp++;

							meta_event = mev;
						}
						else if (event_type == 0x58)
						{
							// Time Signature
							MetaTimeSignature mev = new MetaTimeSignature();

							int numerator = meta_data[mp];
							mev.Numerator = numerator;
							mp++;
							int denominator = meta_data[mp];
							mev.Denominator = (int)Math.Pow(2, denominator);
							mp++;
							int metronome = meta_data[mp];
							mev.ClocksPerClick = metronome;
							mp++;
							int num_32nds = meta_data[mp];
							mev.ThirtySecondNotesPerQuarterNote = num_32nds;
							mp++;

							meta_event = mev;
						}
						else if (event_type == 0x59)
						{
							// Key Signature
							MetaKeySignature mev = new MetaKeySignature();

							int keys = meta_data[mp]; // -7 to 7
							mev.Keys = keys;
							mp++;
							if (meta_data[mp] == 0) 
							{
								mev.IsMajorScale = true;
							}
							else
							{
								mev.IsMinorScale = true;
							}
							mp++;

							meta_event = mev;
						}
						else if (event_type == 0x7F)
						{
							// Sequencer Specific
							MetaSequencerSpecific mev = new MetaSequencerSpecific();
							mev.Data = meta_data;
							meta_event = mev;
						}

						ev.MetaEvent = meta_event;
					}
					else
					{
						// TODO: error! - unknown status byte
						throw ( new Exception("Unknown status byte!") );
					}
					tracks[i].addEvent(ev);
				}
			}
		}

		// nkmidi.Sequence has similar method
		private uint convertToUIntFromBytes(byte[] byte_array, int start_index, int length)
		{
			byte[] tmp;

			if (length == 1 || length == 3) 
			{
				tmp = new byte[length+1];
				tmp[0] = 0;
				Array.Copy(byte_array, start_index, tmp, 1, length);
			}
			else
			{
				tmp = new byte[length];
				Array.Copy(byte_array, start_index, tmp, 0, length);
			}

			Array.Reverse(tmp);

			if (length == 1 || length == 2) 
			{
				return BitConverter.ToUInt16(tmp, 0);
			} 
			else if (length == 3 || length == 4) 
			{
				return BitConverter.ToUInt32(tmp, 0);
			}

			// TODO
			throw ( new ArgumentException() );
		}

		private uint convertVariableLengthValue(byte[] byte_array, int start_index, out int bytes_processed)
		{
			uint val = 0;
			int p = start_index;

			while ( (byte_array[p] & 0x80 ) == 0x80 )
			{
				// the first bit of the byte is 1, so add the rest 7 bits and continue
				val = (val << 7);
				val += (uint)(byte_array[p] & 0x7F);
				p++;
			}
			// the first bit of the byte is 0, so this is the last 7-bit fragment of the true value
			val = (val << 7);
			val += (uint)(byte_array[p] & 0x7F);
			p++;

			bytes_processed = p - start_index; // how many bytes are processed

			return val;
		}

		// create a new array of bytes
		private byte[] copyBytes(byte[] byte_array, int start_index, int length)
		{
			byte[] newArray = new byte[length];
			Array.Copy(byte_array, start_index, newArray, 0, length);
			return newArray;
		}

		// create a new array of bytes (with leading byte)
		private byte[] copyBytesWithLeadingByte(byte[] byte_array, int start_index, int length, byte leading_byte)
		{
			byte[] newArray = new byte[length+1];
			Array.Copy(byte_array, start_index, newArray, 1, length);
			newArray[0] = leading_byte;
			return newArray;
		}



		// accessors and properties

		public string HeaderChunkType
		{
			get { return headerChunkType; }
			set { headerChunkType = value; }
		}

		public uint HeaderDataLength
		{
			get { return headerDataLength; }
			set { headerDataLength = value; }
		}

		public uint FormatNumber
		{
			get { return formatNumber; }
			set { formatNumber = value; }
		}

		public uint NumberOfTracks
		{
			get { return numberOfTracks; }
			set { numberOfTracks = value; }
		}

		public uint TimeBase
		{
			get { return timeBase; }
			set { timeBase = value; }
		}

		public MIDITrack[] Tracks
		{
			get { return tracks; }
		}

		public MIDITrack getTrack(int index)
		{
			return tracks[index];
		}
	}
}
